
#include "sys.h"
#include <stdlib.h>

#ifndef WINDOWS
	#include <sys/wait.h>
	#include <errno.h>
	#include <unistd.h>
	#include <signal.h>
	#include <fcntl.h>
	#include <time.h>
#endif

typedef struct
{
	int	running;	// The process is successfully started
	int	exitCode;	// The exit code of the process
	int	eof[2];	// flag EOF for stdout (eof[0]) and stderr (eof[1])

	#ifdef WINDOWS
		int	async;
		PROCESS_INFORMATION	procInfo;
		HANDLE	stdin;	// Sends the input to the child process through this handle
		HANDLE	stdout;	// Reads the output of the child process through this handle
		HANDLE	processStdin;	// The child process reads stdin through this handle
		HANDLE	processStdout;	// The child process writes stdout through this handle
	#else
		pid_t	pid;
		int	stdin;
		int	stdout;
		int	stderr;
	#endif

} SysProcess;

//[of]:start
#ifdef WINDOWS
	static BOOL createPipe(LPHANDLE readPipe, LPHANDLE writePipe)
	{	
		SECURITY_ATTRIBUTES sa;
		sa.nLength = sizeof(SECURITY_ATTRIBUTES);
		sa.bInheritHandle = TRUE;
		sa.lpSecurityDescriptor = NULL;
		
		return CreatePipe (readPipe, writePipe, &sa, 4096);
	}
	
	static SysErrorCode start(	SysProcess*	process, 
		wchar*	commandLine, 
		wchar*	currentDirectory,
		int	showWindow)
	{
		STARTUPINFO si;
		
		// Create stdin redirection
		BOOL success = createPipe(&process->processStdin, &process->stdin);
		if( success == 0 )
			return ERR_SYSTEM;
	
		// Create stdout redirection
		success = createPipe(&process->stdout, &process->processStdout);
		if( success == 0 )
			return ERR_SYSTEM;
	
		// Initialize the startup info structure
		si.cb	= sizeof(STARTUPINFO);
		si.dwFlags	= STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
		si.hStdInput	= process->processStdin;
		si.hStdOutput	= process->processStdout;
		si.hStdError	= process->processStdout;
		si.lpReserved	= NULL;
		si.lpDesktop	= NULL;
		si.lpTitle	= NULL;
		si.dwX	= 0;
		si.dwY	= 0;
		si.dwXSize	= 0;
		si.dwYSize	= 0;
		si.dwXCountChars	= 0;
		si.dwYCountChars	= 0;
		si.dwFillAttribute	= 0;
		si.wShowWindow	= showWindow ? SW_SHOW : SW_HIDE;
		si.cbReserved2	= 0;
		si.lpReserved2	= NULL;
	
		success = CreateProcess(	NULL, 
			commandLine,
			NULL,	// process security
			NULL,	// thread security
			TRUE,
			CREATE_NEW_PROCESS_GROUP,
			NULL,
			(*currentDirectory==0 ? NULL : currentDirectory),
			&si,
			&process->procInfo);
	
		// We don't read from the stdin pipe, only the child process is using this part
		CloseHandle(process->processStdin);
		// We don't send anything to the child today
		CloseHandle(process->stdin);
	
		if( success == 0 )
			return ERR_SYSTEM;
		
		return OK;
	}

#else
	static SysErrorCode start(	SysProcess*	process, 
		char*	commandLine, 
		char*	directory,
		int	async)
	{
		int	pid;
		int	stdin[2];
		int	stdout[2];
		int	stderr[2];
		char*	args[4];
		
		if( pipe(stdin) != 0 ) return 0;
		if( pipe(stdout) != 0 ) return 0;
		
		// Stderr will use same pipe as stout
		stderr[0] = dup (stdout[0]);
		stderr[1] = dup (stdout[1]);
		
		pid = fork();
		
		// Fork failed ?
		if( pid == -1 )
			return ERR_SYSTEM;
		
		// Child process ?
		if( pid == 0 )
		{
			int handle; // unused, only exist to make gcc happy
			
			// The stdin-pipe input becomes stdin
			close(0);
			handle = dup(stdin[0]);
			close(stdin[0]);
			close(stdin[1]);
			
			// The stdout-pipe output becomes stdout
			close(1);
			handle = dup(stdout[1]);
			close(stdout[0]);
			close(stdout[1]);
			
			// The stderr-pipe output becomes stderr
			close(2);
			handle = dup(stderr[1]);
			close(stderr[0]);
			close(stderr[1]);
			
			// Empty directory means same as parent process
			if( directory[0] != 0 )
			{
				int res = chdir(directory);
				if( res != 0 )
					_exit (127);
			}
			
			args[0]	= "sh";
			args[1]	= "-c";
			args[2]	= commandLine;
			args[3]	= 0;
			execv("/bin/sh", args);
			_exit (127);
		}
	
		close(stdin[0]);
		close(stdin[1]);
		close(stderr[1]);
		close(stdout[1]);
	
		if( async )
		{
		  fcntl(stdout[0], F_SETFL, O_NONBLOCK);
		  fcntl(stderr[0], F_SETFL, O_NONBLOCK);
		}
	
		process->pid	= pid;
		process->stdout	= stdout[0];
		process->stderr	= stderr[0];
		return OK;
	}
#endif
//[cf]
//[of]:processIsRunning
ENTRY int csys_processIsRunning(SysProcess* process)
{
	if( process->running )
	{	
		#ifdef WINDOWS
			// Get exit code
			DWORD exitCode;
			GetExitCodeProcess(process->procInfo.hProcess, &exitCode);
		
			// Still active ?
			if( exitCode != STILL_ACTIVE )
			{
				process->exitCode = (int)exitCode;
				process->running = 0;
			}
		#else
			int status;
			int res = waitpid(process->pid, &status, WNOHANG);
			if( res == -1 )
			{
				// The only possible error here should be EINTR:
				// the process is still running, we just have been interrupted.
			}
			else if( res != 0 )
			{
				process->exitCode = WEXITSTATUS(status);
				process->running = 0;
			}
		#endif
	}
	return process->running ? 1 : 0;
}
//[cf]
//[of]:processIsEof
//[c]Returns true if the end of file is reached for the given channel
//[c]
ENTRY int csys_processIsEof(SysProcess* process, int channel)
{
	return process->eof[channel];
}
//[cf]
//[of]:processReadPipe
//[c]Read data from the pipe.
//[c]
//[c]PARAMETERS
//[c]	process	--	The running process.
//[c]	channel	--	0 = stdout, 1 = stderr.
//[c]	buffer	--	The buffer to fill.
//[c]	bufferSize	--	The size of the buffer to fill. It must be > 0.
//[c]	out	--	[out] The number of bytes written in the buffer.
//[c]
//[c]RETURN VALUES
//[c]	An error code.
//[c]	
ENTRY SysErrorCode csys_processReadPipe(SysProcess* process, int channel, unsigned char* buffer, size_t bufferSize, size_t* out)
{
	#ifdef WINDOWS
		DWORD read;
		BOOL success;
		HANDLE handle;
		
		// stderr ignored
		handle = process->stdout;
		
		if( process->async )
		{
			// Check if the pipe is empty or not
			DWORD available;
			success = PeekNamedPipe (handle, NULL, 0, NULL, &available, NULL);
			if( success == 0 )
				return ERR_SYSTEM;
			
			// No byte available
			if( available == 0 )
			{
				*out = 0;
				return OK;
			}
		}
		
		// Read bytes
		success = ReadFile(handle, buffer, bufferSize, &read, NULL);
		if( success == 0 )
			return ERR_SYSTEM;
		*out = read;
		process->eof[channel] = (read == 0);
	#else
		int rd;
		int fd = channel ? process->stderr : process->stdout;
		
		// Read bytes
		rd = read(fd, buffer, bufferSize);
		if( rd == -1 )
		{
			if( errno == EAGAIN )
				*out = 0;
			else
				return ERR_SYSTEM;
		}
		else
		{
			*out = rd;
			process->eof[channel] = (rd == 0);
		}
	#endif

	return OK;
}
//[cf]
//[of]:processExitCode
ENTRY int csys_processExitCode(SysProcess* process)
{
	return process->exitCode;
}
//[cf]
//[of]:processStop
ENTRY SysErrorCode csys_processStop(SysProcess* process)
{
	#ifdef WINDOWS
		BOOL success = TerminateProcess (process->procInfo.hProcess, 1);
		return ( success != 0 ) ? OK : ERR_SYSTEM;
	#else
		int res = kill(process->pid, SIGTERM);
		return (res == 0) ? OK : ERR_SYSTEM;
	#endif
}
//[cf]
//[of]:processWait
ENTRY SysErrorCode csys_processWait(SysProcess* process)
{
	if( process->running )
	{
		#ifdef WINDOWS
	
			DWORD exitCode;
			BOOL success;
			
			DWORD res = WaitForSingleObject (process->procInfo.hProcess, INFINITE);
			if( res == WAIT_FAILED )
				return ERR_SYSTEM;
			
			success = GetExitCodeProcess(process->procInfo.hProcess, &exitCode);
			if( success == 0 )
				return ERR_SYSTEM;
		
			process->exitCode = (int)exitCode;
			
		#else
	
			int status;
			int res = waitpid(process->pid, &status, 0);
			if( res == -1 )
				return ERR_SYSTEM;
			process->exitCode = WEXITSTATUS(status);
	
		#endif
		
		process->running = 0;
	}
	return OK;
}
//[cf]
//[of]:processDelete
ENTRY void csys_processDelete(SysProcess* process)
{
	#ifdef WINDOWS
		if( process->stdout != NULL )
			CloseHandle(process->stdout);
		if( process->processStdout != NULL )
			CloseHandle(process->processStdout);
		
		if( process->procInfo.hProcess != NULL )
			CloseHandle(process->procInfo.hProcess);
	
		if( process->procInfo.hThread != NULL )
			CloseHandle(process->procInfo.hThread);
	#else
		if( process->stdin != -1 )
			close(process->stdin);
		
		if( process->stdout != -1 )
			close(process->stdout);
		
		if( process->stderr != -1 )
			close(process->stderr);
		
	#endif
	csys_free(process, sizeof(SysProcess));
}
//[cf]
//[of]:processNew
ENTRY SysErrorCode csys_processNew(	String	commandLine, 
	String	currentDirectory,
	int	showWindow,
	int	async,
	SysProcess**	out)
{
	SysErrorCode res;
	SysProcess* process;
	
	#ifdef WINDOWS
		wchar* wideCommandLine;
		wchar* wideCurrentDirectory;
		
		process = (SysProcess*)csys_allocate(sizeof(SysProcess));
		
		// initialize structure
		process->running	= 0;
		process->async	= async;
		process->exitCode	= 0;
		process->stdin	= NULL;
		process->stdout	= NULL;
		process->processStdin	= NULL;
		process->processStdout	= NULL;
		process->procInfo.hProcess	= NULL;
		process->procInfo.hThread	= NULL;
		process->eof[0]	= 0;
		process->eof[1]	= 0;

		res = csys_newWideFromString(&wideCommandLine, commandLine);
		if( res == OK )
		{
			res = csys_newWideFromString(&wideCurrentDirectory, currentDirectory);
			if( res == OK )
			{
				// create the process and start it
				res = start(process, wideCommandLine, wideCurrentDirectory, showWindow);
				if( res == OK )
					process->running = 1;
				
				csys_deleteWide(wideCurrentDirectory);
			}
			csys_deleteWide(wideCommandLine);
		}			
	#else
		char* localeCommandLine;
		char* localeCurrentDirectory;
		
		process = (SysProcess*)csys_allocate(sizeof(SysProcess));
		
		// Initialize
		process->running	= 0;
		process->pid	= 0;
		process->exitCode	= 0;
		process->stdin	= -1;
		process->stdout	= -1;
		process->stderr	= -1;
		process->eof[0]	= 0;
		process->eof[1]	= 0;
		
		res = csys_newLocaleFromString(&localeCommandLine, commandLine);
		if( res == OK )
		{
			res = csys_newLocaleFromString(&localeCurrentDirectory, currentDirectory);
			if( res == OK )
			{
				// Create the process and start it
				res = start(process, localeCommandLine, localeCurrentDirectory, async);
				if( res == OK )
					process->running	= 1;
			
				csys_deleteLocale(localeCurrentDirectory);
			}
			csys_deleteLocale(localeCommandLine);
		}
	#endif

	if( res != OK )
		csys_processDelete(process);
	else
		*out = process;
	return res;
}
//[cf]

//[of]:sleep
ENTRY void csys_sleep(int ms)
{
	#ifdef WINDOWS
	    Sleep (ms);
	#else
		struct timespec req;
		struct timespec rem;
		req.tv_sec = ms / 1000;
		req.tv_nsec = (long)(ms % 1000)* 1000000;
		nanosleep(&req, &rem);
	#endif
}
//[cf]
//[of]:atexit
ENTRY void csys_atexit(void (*function)(void))
{
	atexit(function);
}
//[cf]
